from .__about__ import __version__
from .api import check_har, check_perf_log
from .models import Result, TestCase, TrackingPlan

from .exceptions import (
    GAUnitException,
    TrackingPlanError,
    DictXORJsonPathError,
    TestCaseCheckError,
)



def generate_mock_har(*args) -> dict:
    """build facility for unit tests used to generate hars with GA hits

    GA hits will get values in argument

    Args:
        schema (list): 'dp' parameters for GA hits "A" ,"x", "B",..

    Returns :
        dict : har
        example : generate_mock_har_ga("A","B") ->
            ({'log': {'entries': [
                {'request': {'url': 'https://www.google-analytics.com/collect?v=1&dp=A'}},
                {'request': {'url': 'https://www.google-analytics.com/collect?v=1&dp=B'}}
            ]}}
    """

    ga_urls = [
        "https://www.google-analytics.com/collect?v=1&dp=" + dp for dp in [*args]
    ]
    har = {
        "log": {
            "entries": [{"request": {"method": "GET", "url": url}} for url in ga_urls]
        }
    }
    return har


def generate_mock_perf_log(*args) -> list:
    """build facility for unit test used to generate webdriver Performance Log

    GA hits will get values in argument

    Returns:
        list: performance log
        example : generate_mock_perf_log_ga("A") ->
            [{
                "level": "INFO",
                "message": "{\"message\":{\"method\":\"Network.requestWillBeSent\",\"params\":{\"request\":"
                    +"{\"url\":\"https://www.google-analytics.com/collect?v=1&t=pageview&dp=dp=A\""
                    + "}}}}"

            }]
    """
    # fmt: off
    perf_log = [
        {
            "level": "INFO",
            "message": "{\"message\":{\"method\":\"Network.requestWillBeSent\",\"params\":{\"request\":{\"url\":\"" 
                + "https://www.google-analytics.com/collect?v=1&t=pageview&dp=" 
                +  dp
                + "\"}}}}"
        }
        for dp in [*args]
    ]
    # fmt: on
    return perf_log



import pprint
import sys

import click

import gaunit
from gaunit.utils import (
    filter_keys,
    get_ga_requests_from_har,
    open_json,
    parse_ga_request,
)

from .__about__ import __version__


@click.group(context_settings={"help_option_names": ["-h", "--help", "help"]})
@click.version_option(
    prog_name="GAUnit", version=__version__, message="%(prog)s %(version)s"
)
@click.pass_context
def cli(context):
    pass


@click.command(help="Print this help")
def help():
    with click.Context(cli) as context:
        click.echo(cli.get_help(context))


@click.command("check", help="Check events against an existing tracking plan")
@click.argument("har_file", type=click.Path())
@click.argument("test_case")
@click.option(
    "-t",
    "--tracking_plan",
    type=click.Path(),
    default="./tracking_plan.json",
)
@click.option(
    "-a",
    "--all",
    is_flag=True,
    help="print all expected events (missing and found)",
)
def check(test_case, har_file, tracking_plan, all):
    # TODO : test_case should be optionnal if tracking plan has only one test_case
    # if args.tracking_plan:
    #  ..
    # else:
    #     if 1 test case in tracking plan
    #         r = gaunit.check_har(args.test_case, args.tracking_plan, har_path=args.har_file)
    #     # N test case
    #     else:
    #     # # error
    #     # print("error: more than one test case in tracking plan, please specify a '--test-case' parameter ")
    #     # pass

    tp = gaunit.TrackingPlan.from_json(tracking_plan)
    r = gaunit.check_har(test_case, tracking_plan=tp, har_path=har_file)

    r.print_result(display_ok=all)
    if False in r.checklist_expected_events:
        sys.exit(1)  # end with return code 1 if check failed


@click.command("extract", help="From a HAR file, extract and print GA events")
@click.argument("har_file")
@click.option(
    "-f",
    "--filter",
    help="list of specific events parameters to extract seperated by `,` (other params are filtered out). Example: '--filter a,b,c'",
)
def extract(har_file, filter):

    har = open_json(har_file)
    requests = get_ga_requests_from_har(har)
    events = []
    for r in requests:
        events.extend(parse_ga_request(r))

    if filter:
        param_filter = filter.split(",")
        events = [filter_keys(e, param_filter) for e in events]
    pprint.pprint(events)


cli.add_command(help)
cli.add_command(check)
cli.add_command(extract)

if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



class GAUnitException(Exception):
    """There was an exception while running GAUnit"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TrackingPlanError(GAUnitException):
    """Failed to instantiate :class:`~gaunit.TrackingPlan` or :class:`~gaunit.TrackingPlan` is invalid"""


class DictXORJsonPathError(GAUnitException):
    """One and only one argument must be provided: dict or JSON path"""


class TestCaseCheckError(GAUnitException):
    """Something is missing in Test Case to make a proper check"""



import unittest

import gaunit


class test_utils(unittest.TestCase):
    def test_get_ga_requests_from_har_OK(self):
        har = {
            "log": {
                "entries": [
                    {"request": {"url": "https://domain.com"}},
                    {"request": {"url": "https://www.google-analytics.com/collect"}},
                    {"request": {"url": "https://domain.com"}},
                ]
            }
        }
        requests = gaunit.utils.get_ga_requests_from_har(har)
        self.assertEqual(
            [{"url": "https://www.google-analytics.com/collect"}], requests
        )

    # other test_get_ga_requests_from_har ?

    def test_parse_ga_request_GET(self):
        request = {
            "method": "GET",
            "url": "https://www.google-analytics.com/g/collect?v=1",
        }
        events = gaunit.utils.parse_ga_request(request)
        self.assertEqual([{"v": "1"}], events)

    def test_parse_ga_request_POST(self):
        request = {
            "method": "POST",
            "url": "https://www.google-analytics.com/g/collect?v=2",
            "postData": {"text": "en=page_view\r\nen=scroll"},
        }
        events = gaunit.utils.parse_ga_request(request)
        self.assertEqual(
            [{"v": "2", "en": "page_view"}, {"v": "2", "en": "scroll"}], events
        )

    def test_parse_ga_request_GET_url_encode(self):
        request = {
            "method": "GET",
            "url": "https://www.google-analytics.com/g/collect?dp=%2F",
        }
        events = gaunit.utils.parse_ga_request(request)
        self.assertEqual([{"dp": "/"}], events)

    def test_parse_ga_request_POST_url_encode(self):
        request = {
            "method": "POST",
            "url": "https://www.google-analytics.com/g/collect?v=2",
            "postData": {"text": "en=page_view&dt=%2F"},
        }
        events = gaunit.utils.parse_ga_request(request)
        self.assertEqual([{"v": "2", "en": "page_view", "dt": "/"}], events)

    def test_parse_ga_request_GET_numbers(self):
        request = {
            "method": "GET",
            "url": "https://www.google-analytics.com/g/collect?ev=1&pr1pr=2.0",
        }
        events = gaunit.utils.parse_ga_request(request)
        self.assertEqual([{"ev": "1", "pr1pr": "2.0"}], events)

    def test_parse_ga_request_POST_numbers(self):
        request = {
            "method": "POST",
            "url": "https://www.google-analytics.com/g/collect?v=2",
            "postData": {"text": "ev=1&pr1pr=2.0"},
        }
        events = gaunit.utils.parse_ga_request(request)
        self.assertEqual([{"v": "2", "ev": "1", "pr1pr": "2.0"}], events)

    def test_parse_ga_request_POST_no_postData(self):
        request = {
            "method": "POST",
            "url": "https://www.google-analytics.com/g/collect?v=2",
        }
        events = gaunit.utils.parse_ga_request(request)
        self.assertEqual([{"v": "2"}], events)

    def test_parse_postdata_events_1(self):
        data = "p=v"
        events = gaunit.utils.parse_postdata_events(data)
        self.assertEqual([{"p": "v"}], events)

    def test_parse_postdata_events_2(self):
        data = "p=v\r\np=v"
        events = gaunit.utils.parse_postdata_events(data)
        self.assertEqual([{"p": "v"}, {"p": "v"}], events)

    def test_parse_postdata_events_3(self):
        data = "p=v\r\np=v&pp=vv"
        events = gaunit.utils.parse_postdata_events(data)
        self.assertEqual([{"p": "v"}, {"p": "v", "pp": "vv"}], events)

    def test_parse_postdata_events_empty(self):
        data = ""
        events = gaunit.utils.parse_postdata_events(data)
        self.assertEqual([{}], events)

    def test_get_ga_requests_from_browser_perf_log_OK(self):
        # fmt: off
        perf_log = [
            # log entry we want to get
            {
                "level": "INFO",
                "message": "{\"message\":{\"method\":\"Network.requestWillBeSent\",\"params\":{\"request\":{\"url\":\"https://www.google-analytics.com/collect\"}}}}", 
            },
            # skipped
                        {
                "level": "INFO",
                "message": "{\"message\":{\"method\":\"Network.responseReceived\",\"params\":{}}}", 
            }
        ]
        # fmt: on
        urls = gaunit.utils.get_ga_requests_from_browser_perf_log(perf_log)
        self.assertEqual(["https://www.google-analytics.com/collect"], urls)

    def test_load_dict_xor_json_too_many_args(self):
        d = {"dummy": "dummy"}
        json_path = "dummy.json"
        with self.assertRaises(gaunit.exceptions.DictXORJsonPathError):
            gaunit.utils.load_dict_xor_json(d, json_path)

    def test_load_dict_xor_json_no_args(self):
        d = None
        json_path = None
        with self.assertRaises(gaunit.exceptions.DictXORJsonPathError):
            gaunit.utils.load_dict_xor_json(d, json_path)

    # def get_ga_request_type(self):
    #     urls = [
    #         "https://www.google-analytics.com/collect?v=1..",
    #         "https://www.google-analytics.com/j/collect?v=1..",
    #         "https://domain.com",
    #         "https://analytics.google.com/g/collect?v=2...",
    #     ]
    #     ga_urls = gaunit.utils.filter_ga_urls(urls)
    #     self.assertEqual(
    #         [
    #             "https://www.google-analytics.com/collect?v=1..",
    #             "https://www.google-analytics.com/j/collect?v=1..",
    #         ],
    #         ga_urls,
    #     )

    def test_parse_url_params(self):
        url = "https://www.google-analytics.com/collect?v=1&t=pageviews&dp=home"
        params = gaunit.utils.parse_ga_url(url)
        self.assertEqual({"v": "1", "t": "pageviews", "dp": "home"}, params)
        # TODO parse url without params : {}

    def test_is_ga_url(self):
        self.assertTrue(
            gaunit.utils.is_ga_url("https://www.google-analytics.com/collect")
        )


if __name__ == "__main__":
    unittest.main()



"""
gaunit.api

This modules makes gaunit nicer to use
"""
from .models import Result, TestCase, TrackingPlan


def check_har(
    test_case_id: str, tracking_plan: TrackingPlan, har=None, har_path=None
) -> Result:
    """Performs checks of a har dict or HAR JSON file against a :class:`~gaunit.TrackingPlan`.

    Example:
        >>> expected_events = [{"t":"pageview","dt":"home"},...]
        >>> tracking_plan = gaunit.TrackingPlan.from_events("my_test_case", expected_events)
        >>> r = gaunit.check_har("my_test_case", tracking_plan, har=har) # from har dict
        >>> r.was_sucessful()
        True
        >>> r = gaunit.check_har("my_test_case", tracking_plan, har_path="har.json") # from HAR file
        >>> r.was_sucessful()
        True

    Args:
        test_case_id (str): test case id (same id used to match with tracking plan)
        tracking_plan (:class:`~gaunit.TrackingPlan`): tracking plan containing expected events for this
            test case. Defaults to None
        har (dict): actual har for this test case in dict format. Defaults to None
        har_path (str) : path to HAR file for this test case (standard HAR JSON). Defaults to None

    Note:
        One and one only argument must be given: ``har`` or ``har_path``

    Returns:
        :class:`gaunit.Result`: complete results of your test case.
    """
    tc = TestCase(test_case_id, tracking_plan=tracking_plan, har=har, har_path=har_path)
    return tc.result()


def check_perf_log(
    test_case_id: str, tracking_plan: TrackingPlan, perf_log: list
) -> Result:
    """Performs checks of a Performance log against a :class:`~gaunit.TrackingPlan`.


    Warning:
        Performance logs do not record POST requests (massively used with GA4 and soon, according to
        Google, with gtag.js). This method will be
        deprecated in future versions. We advise you to use :func:`gaunit.check_har` instead

    Example:
        See full example on Github
        `here <https://github.com/VinceCabs/GAUnit/tree/master/examples/auto_test_with_perf_log>`_.

    Args:
        test_case_id (str): test case id (same id used to match with tracking plan)
        tracking_plan (:class:`~gaunit.TrackingPlan`): tracking plan containing expected events for this
            test case. Defaults to None
        perf_log (list): log entries from driver

    Returns:
        :class:`gaunit.Result`: complete results of your test case.
    """
    tc = TestCase(test_case_id, tracking_plan=tracking_plan, perf_log=perf_log)
    return tc.result()



import unittest

import gaunit

from tests.utils import generate_mock_har, generate_mock_perf_log


class test_api(unittest.TestCase):
    def setUp(self) -> None:
        self.tp = gaunit.TrackingPlan()
        self.tp.update_test_case("home_engie", [{"dp": "A"}, {"dp": "B"}, {"dp": "C"}])

    def test_check_har(self):

        har = generate_mock_har("A", "B", "C")
        r = gaunit.check_har("home_engie", self.tp, har=har)
        self.assertEqual([True, True, True], r.checklist_expected_events)

    def test_check_perf_log(self):

        perf_log = generate_mock_perf_log("A", "B", "C")
        r = gaunit.check_perf_log("home_engie", self.tp, perf_log)
        self.assertEqual([True, True, True], r.checklist_expected_events)


if __name__ == "__main__":
    unittest.main()



""""
Automatic test with Perf Logs

Code sample: an "add to cart" scenario using Selenium and Chrome Performance log to 
automatically test tracking on Google online store demo.
Warning : this method will not work with GA4
"""
from os.path import abspath, dirname, join
from time import sleep

import gaunit
from selenium import webdriver
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities


def run():

    # set up webdriver
    capabilities = DesiredCapabilities.CHROME
    capabilities["goog:loggingPrefs"] = {"performance": "ALL"}  # chromedriver 75+
    driver = webdriver.Chrome(desired_capabilities=capabilities)

    # start test case
    driver.implicitly_wait(10)
    driver.get("https://enhancedecommerce.appspot.com/")
    sleep(2)
    driver.find_element_by_id("homepage-9bdd2-1").click()
    sleep(2)
    driver.find_element_by_id("addToCart").click()
    sleep(2)
    # Get Performance Log and close all
    log = driver.get_log("performance")
    driver.quit()

    # # Uncomment if you want to store log in json file
    # with open("perf_log.json", "w", encoding="utf8") as f:
    #     json.dump(log, f, indent=2)

    # # Uncomment if you want to parse existing log file before check
    # with open("perf_log.json", encoding="utf8") as f:
    #     log = json.load(f)

    # check hits against tracking plan and print results
    path = join(abspath(dirname(__file__)), "tracking_plan.json")
    tracking_plan = gaunit.TrackingPlan.from_json(path)
    r = gaunit.check_perf_log("ga_demo_store_add_to_cart", tracking_plan, log)

    r.print_result(display_ok=True)


if __name__ == "__main__":
    run()



""""
Open a manual test session

Code sample: opens a new browser session and records a har. At the end of the session, 
it checks the har against tracking plan. Tracking plan is the "add to cart" scenario on 
Google online store demo described in GAUnit Documentation: 
https://gaunit.readthedocs.io/en/latest/getting_started.html
"""
import json
from os.path import abspath, dirname, join
from time import sleep
from tkinter import messagebox


import gaunit
from browsermobproxy import Server
from selenium import webdriver


def run():

    # set up proxy
    server = Server()  # or add path to binary: 'Server(path="browsermob-proxy")'
    server.start()
    # 'useEcc' is needed to have decent response time with HTTPS
    proxy = server.create_proxy({"useEcc": True})

    # set up Geckodriver/Firefox
    # # uncomment if you want to use Firefox
    # profile = webdriver.FirefoxProfile()
    # profile.set_proxy(proxy.selenium_proxy())
    # driver = webdriver.Firefox(firefox_profile=profile)

    # set up Chrome driver
    options = webdriver.ChromeOptions()
    options.add_argument("--proxy-server=%s" % proxy.proxy)
    # options.add_argument("--headless")  # uncomment if you want headless Chrome
    capabilities = webdriver.DesiredCapabilities.CHROME.copy()
    capabilities["acceptInsecureCerts"] = True
    driver = webdriver.Chrome(options=options, desired_capabilities=capabilities)

    # start test case
    driver.implicitly_wait(10)
    test_case = "demo_store_add_to_cart"
    # 'captureContent' for POST requests
    proxy.new_har(test_case, options={"captureContent": True})
    driver.get("https://enhancedecommerce.appspot.com/")

    messagebox.showinfo(
        "Manual browsing mode",
        "Recording network trafic. Browse site, then press 'OK' when you're finished",
    )

    # export har and close all
    har = proxy.har
    server.stop()
    driver.quit()

    # uncomment if you need to export the har
    # with open(
    #     join(abspath(dirname(__file__)), test_case) + ".har", "w", encoding="utf8"
    # ) as f:
    #     json.dump(har, f)

    # check events against tracking plan and print results
    path = join(abspath(dirname(__file__)), "tracking_plan.json")
    tracking_plan = gaunit.TrackingPlan.from_json(path)
    r = gaunit.check_har(test_case, tracking_plan, har=har)

    r.print_result(display_ok=True)


if __name__ == "__main__":
    run()



""""
Automatic test with proxy

Code sample: an "add to cart" scenario using Selenium and a proxy to automatically test 
tracking on Google online store demo.
"""
import json
from os.path import abspath, dirname, join
from time import sleep

import gaunit
from browsermobproxy import Server
from selenium import webdriver


def run():

    # set up proxy
    server = Server()  # or add path to binary: 'Server(path="browsermob-proxy")'
    server.start()
    # 'useEcc' is needed to have decent response time with HTTPS
    proxy = server.create_proxy({"useEcc": True})

    # set up Geckodriver/Firefox
    # # uncomment if you want to use Firefox
    # profile = webdriver.FirefoxProfile()
    # profile.set_proxy(proxy.selenium_proxy())
    # driver = webdriver.Firefox(firefox_profile=profile)

    # set up Chrome driver
    options = webdriver.ChromeOptions()
    options.add_argument("--proxy-server=%s" % proxy.proxy)
    # options.add_argument("--headless")  # uncomment if you want headless Chrome
    capabilities = webdriver.DesiredCapabilities.CHROME.copy()
    capabilities["acceptInsecureCerts"] = True
    driver = webdriver.Chrome(options=options, desired_capabilities=capabilities)

    # start test case
    driver.implicitly_wait(10)
    test_case = "ga_demo_store_add_to_cart"
    # 'captureContent' for POST requests
    proxy.new_har(test_case, options={"captureContent": True})
    driver.get("https://enhancedecommerce.appspot.com/")
    sleep(2)
    driver.find_element_by_id("homepage-9bdd2-1").click()
    sleep(2)
    driver.find_element_by_id("addToCart").click()
    sleep(2)

    # export har and close all
    har = proxy.har
    server.stop()
    driver.quit()

    # uncomment if you need to export the har
    # with open(
    #     join(abspath(dirname(__file__)), test_case) + ".har", "w", encoding="utf8"
    # ) as f:
    #     json.dump(har, f)

    # check hits against tracking plan and print results
    path = join(abspath(dirname(__file__)), "tracking_plan.json")
    tracking_plan = gaunit.TrackingPlan.from_json(path)
    r = gaunit.check_har(test_case, tracking_plan, har=har)

    r.print_result(display_ok=True)


if __name__ == "__main__":
    run()



""""
Getting Started

Code sample: This is the code from the "Getting Started" GAUnit Documentation : 
https://gaunit.readthedocs.io/en/latest/getting_started.html
"""
import gaunit

tracking_plan = gaunit.TrackingPlan.from_json("tracking_plan.json")
r = gaunit.check_har(
    "demo_store_add_to_cart", tracking_plan, har_path="demo_store_add_to_cart.har"
)
print(r.was_successful())
# True



"""
gaunit.models

This module implements main classes used by gaunit: :class:`TrackingPlan, 
:class:`~gaunit.TestCase` and :class:`~gaunit.Result`. 
"""
from __future__ import annotations

import json
import pprint
from typing import List, Tuple

from colorama import Fore, init
from gspread import Spreadsheet

from .exceptions import TestCaseCheckError, TrackingPlanError
from .utils import (
    format_events,
    get_ga_requests_from_browser_perf_log,
    get_ga_requests_from_har,
    get_py_version,
    load_dict_xor_json,
    open_json,
    parse_ga_request,
    parse_ga_url,
)


class TrackingPlan(object):
    """User-defined class object representing a :class:`~gaunit.TrackingPlan`.

    Used as an input to perform a check or create a :class:`~gaunit.TestCase`.

    Example:
        See Documentation :ref:`getting_started` or :ref:`howtos` for more details.
    """

    def __init__(self):
        self.content = {}

    def get_expected_events(self, test_case_id: str) -> list:
        """Get expected events for a given test case.

        Args:
            test_case_id (str):  test case id

        Raises:
            :exception:`gaunit.TrackingPlanError`: if test case is not found in tracking plan
        """
        try:
            tc = self.content[test_case_id]
        except KeyError:
            raise TrackingPlanError(
                "test case not found in tracking plan: '%s'" % test_case_id
            )
        events = tc["events"]
        return events

    @classmethod
    def from_events(
        cls, test_case_id: str, expected_events: List[dict]
    ) -> TrackingPlan:
        """Creates an instance of :class:`~gaunit.TrackingPlan` from a list of expected events
        (only for one test case)

        Example :
            >>> from gaunit import TrackingPlan
            >>> expected_events = [{"t":"pageview","dt":"home"},...]
            >>> tracking_plan = TrackingPlan.from_events("my_test_case", expected_events)

        Args:
            test_case_id (str): test case id
            expected_events (list[Dict]): expected events for this tests case

        Returns:
            :class:`~gaunit.TrackingPlan` instance
        """
        tp = TrackingPlan()
        tp.add_test_case(test_case_id, expected_events)
        return tp

    @classmethod
    def from_json(cls, path: str) -> TrackingPlan:
        """Creates an instance of :class:`TrackingPlan` from a JSON file

        See Documentation for the JSON file format.

        Example:
            >>> from gaunit import TrackingPlan
            >>> tracking_plan = TrackingPlan.from_json("tracking_plan.json")

        Args:
            path (str): path to JSON file representing the tracking plan

        Raises:
            :exception:`gaunit.TrackingPlanError`: if tracking plan format is not valid

        Returns:
            :class:`~gaunit.TrackingPlan` instance.
        """
        tp = TrackingPlan()
        d = open_json(path)
        try:
            test_cases = d["test_cases"]
            for tc in test_cases:
                events = test_cases[tc]["events"]
                events = format_events(events)
                test_cases[tc]["events"] = events
            tp.content = test_cases
            return tp
        except KeyError:
            # TODO custom Exceptions
            raise TrackingPlanError(
                "Tracking plan is not valid (see Documentation) '%s'" % path
            )

    @classmethod
    def from_spreadsheet(cls, sheet: Spreadsheet) -> TrackingPlan:
        """Creates an instance of :class:`~gaunit.TrackingPlan` from a Google Spreadsheet.

        This method uses gspread to connect to Google Sheets and import test cases and
        expected events. See Documentation for the spreadsheet format.

        Examples:
            >>> import gspread
            >>> from gaunit import TrackingPlan
            >>> gc = gspread.service_account()  # authentication
            >>> sh = gc.open("Example spreadsheet")  # open spreadsheet
            >>> tp = TrackingPlan.from_spreadsheet(sh)  # import tracking plan

        Args:
            sheet (gspread.Spreadsheet): gspread instance of the spreadsheet to import

        Returns:
            :class:`TrackingPlan` instance.
        """
        tp = TrackingPlan()
        worksheets = sheet.worksheets()
        for w in worksheets:
            events = w.get_all_records()
            events = format_events(events)
            tp.add_test_case(w.title, events)
        return tp

    # @classmethod
    # def from_csv(cls, path: str) -> TrackingPlan:
    #     # TODO P1
    #     pass

    # @classmethod
    # def from_array(cls, array: List[list]) -> TrackingPlan:
    #     # First column has the name of the test_case
    #     # each row corresponds to an event
    #     # [[["test_case","dp","v"],["home_engie","home","1"],...]]
    #     # TODO
    #     pass

    def to_json(self, file: str):
        """Exports :class:`~gaunit.TrackingPlan` instance into a JSON file.

        Example:
            >>> from gaunit import TrackingPlan
            >>> expected_events = [{"t":"pageview","dt":"home"},...]
            >>> tracking_plan = gaunit.TrackingPlan.from_events("my_test_case", expected_events)
            >>> tracking_plan.to_json("tracking_plan.json")

        Args:
            file (str): target file
        """

        tracking_plan = {"test_cases": self.content}
        with open(file, "w", encoding="utf8") as f:
            json.dump(tracking_plan, f)

    def add_test_case(self, test_case_id: str, expected_events: List[dict]):
        """Add or update expected events for a given test case.

        Example:

            >>> from gaunit import TrackingPlan
            >>> expected_events = [{"t":"pageview","dt":"home"},...]
            >>> tracking_plan = TrackingPlan()
            >>> tracking_plan.add_test_case("my_test_case", expected_events)


        See also:
            :func:`TrackingPlan.from_events()`

        Args:
            test_case_id (str): id of the test case
            expected_events (List[dict]): list of expected events for the given test case

        Raises:
            TypeError: if format of ``expected_events`` is not valid.
        """
        try:
            expected_events = format_events(expected_events)
            d = {test_case_id: {"events": expected_events}}
            self.content.update(d)
        except AttributeError:
            raise TypeError(
                "Invalid events type: '%s'. Please provide a list of events"
                % expected_events
            )

    def update_test_case(self, test_case_id: str, expected_events: List[dict]):
        """Simple alias for :func:`TrackingPlan.add_test_case()`."""
        self.add_test_case(test_case_id, expected_events)


class TestCase(object):
    """User-defined class object representing a test case.

    Used to get results between runned test case and expected tracking plan.

    Note:
        One and one only argument must be given: ``har`` or ``har_path``

    Example:
        >>> from gaunit import TestCase
        >>> events = [{"t":"pageview","dt":"home"},...]
        >>> tc = TestCase("my_test_case", expected_events=events)
        >>> r = tc.check_har(har=har)  # or tc.check_har(har_path=path) for a HAR file
        >>> r.was_sucessful()
        True

    See also:
        :func:`gaunit.check_har()` is more straight forward to use.

    Attributes:
        id (str): Test case id (same id used to match with tracking plan)
        tracking_plan (:class:`~gaunit.TrackingPlan`): Tracking plan containing expected events for this
            test case. Defaults to None
        har (dict): Actual har for this test case in dict format. Defaults to None
        har_path (str) : Path to HAR file for this test case (standard HAR JSON).
            Defaults to None.
        perf_log (list) : Browser performance log
        actual_events (list) : List of GA events params parsed from HAR or http log
            Each event is represented by a dict of params (same as `expected_events`).
            Example: ``[{"t":"pageview","dt":"home"},...]``
    """

    def __init__(
        self,
        id: str,
        tracking_plan: TrackingPlan,
        har: dict = None,
        har_path: str = None,
        perf_log: list = None,
    ):
        # Default empty dicts/lists for dict/lists params.
        har = {} if har is None else har
        perf_log = [] if perf_log is None else perf_log

        self.id = id  # test case name
        if isinstance(tracking_plan, TrackingPlan):
            self.expected_events = tracking_plan.get_expected_events(self.id)
        else:
            raise TypeError(
                "Invalid tracking plan type: '%s'. Please provide a 'TrackingPlan' instance."
                % tracking_plan
            )
        self.actual_events = []

        self.har = har  # for debug, will be killed soon
        self.perf_log = perf_log
        # self.page_flow = [] # will store urls from pages TODO
        # self.page_flow_ids = [] # will store har ids for pages
        if har or har_path:
            self.load_har(har, har_path)

        if perf_log:
            self.load_perf_log(perf_log)

    def load_har(self, har=None, har_path=None):
        """Extracts and stores analytics events from a har.

        Updates :attr:`actual_events`.
        Takes one and one only argument : dict or path to a json file

        Args:
            har (dict, optional): [description]. Defaults to None.
            har_path (str, optional): [description]. Defaults to None.

        """

        har = load_dict_xor_json(har, har_path)
        # extract GA events
        requests = get_ga_requests_from_har(har)
        events = []
        for r in requests:
            events.extend(parse_ga_request(r))

        self.har = har  # TDO remove attribute
        self.actual_events = events

        # # extract pages (urls and har ids)
        # page_flow = get_pages_from_har(har)
        # page_flow_ids = get_pages_ids_from_har(har)

    def load_perf_log(self, perf_log: list):
        """Extracts and stores analytics events from Performance Log.

        For more info on Performance Log, see https://chromedriver.chromium.org/logging/performance-log

        Warning:
            Performance logs do not record POST requests (massively used with GA4 and soon, according to
            Google, with gtag.js). This method will be
            deprecated in future versions. We advise you to use :func:`gaunit.check_har` or
            :func:`TestCase.load_har` instead.

        Args:
            perf_log (list): log entries from ``driver.get_log("performance")``
        """
        # TODO GA4 check that there are no POST methods, otherwise throw an error or warning
        urls = get_ga_requests_from_browser_perf_log(perf_log)
        events = [parse_ga_url(url) for url in urls]

        self.perf_log = perf_log
        self.actual_events = events

    def check(self, ordered=True) -> Tuple[list, list]:
        # TODO make private?
        """Compares events from tracking plan and from log and returns 2 checklists.

        Compares :attr:`expected_events` and :attr:`actual_events`, which makes sense!

        Args:
            ordered (bool, optional): True if we want hits to respect tracking plan
                order (default behavior)

        Raises:
            :exception:`gaunit.TestCaseCheckError`: if no valid HAR or Perf log were provided before check

        Returns:
            Tuple[list, list]: 2 checklists; First checklist tells which event in tracking plan is
            missing. Second checklist tells which analytics event from test case corresponds
            to an expected event.
        """

        # check if everything needed is defined
        if not self.har and not self.perf_log:
            raise TestCaseCheckError(
                "HAR and Perf log are both missing or empty. Please load one before performing a check"
            )

        # start checking
        expected = self.expected_events
        actual = self.actual_events
        chklst_expected = []
        chklst_actual = [False] * len(actual)
        pos = 0  # last checked hit position
        for t in expected:
            check = False
            for index, hit in enumerate(actual[pos:]):
                if t.items() <= hit.items():
                    # expected event found: all params are there
                    check = True
                    chklst_actual[pos + index] = True
                    if ordered:
                        # update last checked hit position to respect tracking plan order (default)
                        pos += index
                    break
                # expected event is not here
            chklst_expected.append(check)

        return chklst_expected, chklst_actual

    def result(self):
        """Performs tracking checks and return a :class:`Result` instance"""

        expected, actual = self.check()
        r = Result(self, expected, actual)
        return r


class Result(object):
    """Let you store, get or print results from a test case in various forms

    Usually, :class:`Result` will be returned by main API methods or :func:`TestCase.result()` .

    See also :
        :func:`gaunit.check_har()`, :func:`TestCase.result()`

    Attributes:
        expected_events (list): Events from tracking plan
        actual_events (list): Events found during test run
        checklist_expected_events (list): Checklist of events from tracking plan which
            are missing (``False`` if missing)
        checklist_actual_events (list): Checklist of events found in log which corresponds
                to an expected event
    """

    def __init__(
        self, test_case: TestCase, checklist_expected: list, checklist_actual: list
    ):
        self.expected_events = test_case.expected_events
        self.actual_events = test_case.actual_events
        self.checklist_expected_events = checklist_expected
        self.checklist_actual_events = checklist_actual
        # self.comparison = None

    # TODO method to return merged results : comparison of both tracker and hits list

    def was_successful(self) -> bool:
        """returns the result of test case

        Returns:
            bool: True if test case was successful
        """
        return all(self.checklist_expected_events)

    def get_status_expected_events(self) -> list:
        """Returns expected event and their status

        Returns:
            list: list of expected events and their status (``True`` if found in actual events))

        Example:
            >>> from gaunit import TestCase
            >>> r = gaunit.check_har("my_test_case", tracking_plan, har=har)
            >>> r.get_status_expected_events()
            [{'event':{'t':'pageview', 'dp': 'home'}, 'found': True},..]
        """

        events = self.expected_events
        chcklst = self.checklist_expected_events

        return [{"event": h, "found": c} for (h, c) in zip(events, chcklst)]

    def get_status_actual_events(self) -> list:
        """Returns actual events and which ones were expected in tracking plan

        Returns:
            list: list of actual hits

        Example:
            >>> from gaunit import TestCase
            >>> r = gaunit.check_har("my_test_case", "tracking_plan.json", har=har)
            >>> r.get_status_actual_events()
            [{'event:{'t':'pageview', 'dp': 'home'}, 'expected': True}
        """

        events = self.actual_events
        chcklst = self.checklist_actual_events

        return [{"event": h, "expected": c} for (h, c) in zip(events, chcklst)]

    def print_result(self, display_ok=False):
        """Pretty print result of a test.

        Two parts: prints the expected events that are missing by defaults or all expected events
        if ``display_ok=True`` then prints a summary.

        Example:
            >>> from gaunit import TestCase
            >>> r = gaunit.check_har("my_test_case", "tracking_plan.json", har=har)
            >>> r.print_result(display_ok=True)
            events in tracking plan: 3
            ================================================================================
            {'t': 'pageview', 'dt': 'Home'}
                                                                                    ... OK
            ================================================================================
            {'t': 'pageview', 'dt': 'Product View'}
                                                                                    ... OK
            ================================================================================
            {'t': 'event',
            'ec': 'ecommerce',
            'ea': 'add_to_cart',
            'ev': '44',
            'pr1nm': 'Compton T-Shirt',
            'pr1pr': '44.00'}
                                                                                    ... OK
            --------------------------------------------------------------------------------
            GA events found: total:9 / ok:3 / missing:0
            ✔ OK: all expected events found

        Args:
            display_ok (bool, optional): if set to ``True``, print all expected events, not only
                missing events. Defaults to ``False``.
        """
        self._print_expected_events(all=display_ok)
        self._print_summary()

    def _print_expected_events(self, all=False):
        """pretty print list of events from tracking plan

        says which tracker was found in test case  ("OK" or "missing"), used in CLI

        Args:
            all (bool, optional): if False, print only missing events if True, print all.
                Defaults to False.
        """
        expected = self.expected_events
        chcklst = self.checklist_expected_events
        init(autoreset=True)  # colorama

        print("events in tracking plan: %s" % len(expected))
        args = (
            {"sort_dicts": False} if get_py_version() >= (3, 8) else {}
        )  # preserve dict params order when printing (only for Python>=3.8)
        for (event, check) in zip(expected, chcklst):
            if not check:
                print(80 * "=")
                pprint.pprint(event, **args)  # pylint: disable=E1123
                print(67 * " ", " ... " + Fore.RED + "missing")
            elif check and all:
                print(80 * "=")
                pprint.pprint(event, **args)  # pylint: disable=E1123
                print(72 * " ", " ... " + Fore.GREEN + "OK")
            else:
                pass

    def _print_summary(self):
        """print a result summary

        Example:
            output:

            | GA events found: total:4 / ok:3 / missing:0
            | ✔ OK: all expected events found

            or

            | GA events found: total:11 / ok:1 / missing:2
            | ❌ FAILED: events missing
            ""

        """
        expected = self.expected_events
        actual = self.actual_events
        chcklst = self.checklist_expected_events
        total_found, expected_found = len(actual), len([c for c in chcklst if c])
        missing = len(expected) - expected_found

        print(80 * "-")
        if total_found == 0:
            print("no GA events found")
        else:
            print(
                "GA events found: total:%s / ok:%s / missing:%s"
                % (total_found, expected_found, missing)
            )
        try:
            if False in chcklst:
                print("\N{Cross Mark} FAILED: events missing")
            else:
                print("\N{Heavy Check Mark} OK: all expected events found")
        except UnicodeEncodeError:
            pass

    def print_actual_events(self):
        """pretty print list of analytics hits from test case

        says which analytics hit was in tracking plan ("OK" or "skip"), used in CLI
        """
        events = self.actual_events
        chcklst = self.checklist_actual_events

        init(autoreset=True)  # colorama
        args = (
            {"sort_dicts": False} if get_py_version() >= (3, 8) else {}
        )  # preserve dict params order when printing (only for Python>=3.8)
        for (event, check) in zip(events, chcklst):
            print(80 * "=")
            pprint.pprint(event, **args)  # pylint: disable=E1123
            if check:
                print(72 * " ", " ... " + Fore.GREEN + "OK")
            else:
                print(70 * " ", " ... skip")



""""
Tracking Plan from Google Sheet

Code sample: import your tracking plan from Google Sheet. Tracking plan can have several
tests cases (one tab each). See gspread documentation (https://gspread.readthedocs.io/) 
for authentication.
Demo spreadsheet is here: https://docs.google.com/spreadsheets/d/1Kd68s3vLrBqtMDW-PaALZF-5bTm-4J450YbJ3NTbZjQ
"""
import gaunit

import gspread
from os.path import join, dirname, realpath

# authentication. Method is up to you ; see gspread documentation.
gc = gspread.service_account(filename="service_account.json")

# open demo file and import tracking plan for our test case.
gsheet = gc.open_by_key("1Kd68s3vLrBqtMDW-PaALZF-5bTm-4J450YbJ3NTbZjQ")
tracking_plan = gaunit.TrackingPlan.from_spreadsheet(gsheet)
print(
    "expected events:\n%s"
    % tracking_plan.get_expected_events("ga_demo_store_add_to_cart")
)

# check GA events in a HAR file against tracking plan and print results
har_path = join(dirname(realpath(__file__)), "demo_store_add_to_cart.har")
r = gaunit.check_har("ga_demo_store_add_to_cart", tracking_plan, har_path=har_path)
print(r.was_successful())
# TODO fix numbers (param pr1pr)



import unittest
from os.path import dirname, join, realpath

import gaunit

from tests.utils import generate_mock_har, generate_mock_perf_log


class test_TrackingPlan(unittest.TestCase):
    def test_from_json_OK(self):
        here = dirname(realpath(__file__))
        path = join(here, "tracking_plan.json")
        tp = gaunit.TrackingPlan.from_json(path)
        self.assertEqual(
            tp.content.get("home_engie", None).get("events", None),
            [{"dp": "A"}, {"dp": "B"}, {"dp": "C"}],
        )

    # TODO with unittest.mock
    # def test_from_json_wrong_format_1(self):
    #     test_cases = {"dummy": "dummy"}
    #     with self.assertRaises(Exception):
    #         gaunit.TrackingPlan(test_cases=test_cases)

    # def test_from_json_wrong_format_2(self):
    #     test_cases = {"home_engie": {"dummy": "dummy"}}
    #     with self.assertRaises(Exception):
    #         gaunit.TrackingPlan(test_cases=test_cases)

    # def test_from_spreadsheet(self):

    def test_add_test_case_create_OK(self):
        events = [{"dp": "A"}]
        tp = gaunit.TrackingPlan()
        tp.add_test_case("home_engie", events)
        self.assertEqual(tp.content, {"home_engie": {"events": events}})

    def test_update_test_case_OK(self):
        # all events are replaced by new events
        events = [{"dp": "A"}]
        tp = gaunit.TrackingPlan()
        tp.add_test_case("home_engie", events)
        tp.add_test_case("home_engie", [{"dp": "X"}])
        self.assertEqual(tp.content, {"home_engie": {"events": [{"dp": "X"}]}})

    def test_get_expected_events_missing_test_case(self):
        events = []
        tp = gaunit.TrackingPlan()
        tp.add_test_case("not_my_test_case", events)
        with self.assertRaises(gaunit.exceptions.TrackingPlanError):
            tp.get_expected_events("home_engie")

    def test_get_expected_events_OK(self):
        events = [{"t": "pageview"}]
        tp = gaunit.TrackingPlan()
        tp.add_test_case("home_engie", events)
        events = tp.get_expected_events("home_engie")
        self.assertEqual([{"t": "pageview"}], events)

    def test_get_expected_events_with_int_OK(self):
        events = [{"ev": 1}]
        tp = gaunit.TrackingPlan()
        tp.add_test_case("home_engie", events)
        events = tp.get_expected_events("home_engie")
        self.assertEqual([{"ev": "1"}], events)

    def test_get_expected_events_with_float_OK(self):
        events = [{"ev": 1.0}]
        tp = gaunit.TrackingPlan()
        tp.add_test_case("home_engie", events)
        events = tp.get_expected_events("home_engie")
        self.assertEqual([{"ev": "1.0"}], events)

    def test_get_expected_events_with_url_decode_OK(self):
        events = [{"dl": "%2F"}]
        tp = gaunit.TrackingPlan()
        tp.add_test_case("home_engie", events)
        events = tp.get_expected_events("home_engie")
        self.assertEqual([{"dl": "/"}], events)


class test_TestCase(unittest.TestCase):
    def setUp(self) -> None:
        events = [{"dp": "A"}, {"dp": "B"}, {"dp": "C"}]
        tp = gaunit.TrackingPlan()
        tp.add_test_case("home_engie", events)
        events = tp.get_expected_events("home_engie")
        self.tc = gaunit.TestCase("home_engie", tp)

    def test_constructor_tracking_plan_wrong_type(self):
        with self.assertRaises(TypeError):
            gaunit.TestCase("home_engie", "wrong")

    def test_load_har_ok(self):
        har = {
            "log": {
                "entries": [
                    {
                        "request": {
                            "method": "GET",
                            "url": "https://www.google-analytics.com/collect?v=1&t=pageview&dp=A",
                        }
                    }
                ]
            }
        }
        self.tc.load_har(har)
        self.assertEqual(
            self.tc.actual_events,
            [
                {
                    "v": "1",
                    "t": "pageview",
                    "dp": "A",
                }
            ],
        )

    def test_load_perf_log_ok(self):
        perf_log = generate_mock_perf_log("B")
        self.tc.load_perf_log(perf_log)
        self.assertEqual(
            self.tc.actual_events,
            [
                {
                    "v": "1",
                    "t": "pageview",
                    "dp": "B",
                }
            ],
        )

    def test_check_OK(self):
        har = generate_mock_har("A", "B", "C")
        self.tc.load_har(har)
        checklist_expected, checklist_actual = self.tc.check()
        self.assertEqual(checklist_expected, [True, True, True])
        self.assertEqual(checklist_actual, [True, True, True])

    def test_check_missing_1(self):
        """tracking plan is "A", "B", "C" """
        har = generate_mock_har("A", "x", "B", "x")
        self.tc.load_har(har)
        checklist_expected, checklist_actual = self.tc.check()
        self.assertEqual(checklist_expected, [True, True, False])
        self.assertEqual(checklist_actual, [True, False, True, False])

    def test_check_missing_2(self):
        har = generate_mock_har("A", "C", "x")
        self.tc.load_har(har)
        checklist_expected, checklist_actual = self.tc.check()
        self.assertEqual(checklist_expected, [True, False, True])
        self.assertEqual(checklist_actual, [True, True, False])

    def test_check_OK_not_ordered(self):
        har = generate_mock_har("A", "x", "C", "B")
        self.tc.load_har(har)
        checklist_expected, checklist_actual = self.tc.check(ordered=False)
        self.assertEqual(checklist_expected, [True, True, True])
        self.assertEqual(checklist_actual, [True, False, True, True])

    def test_check_missing_not_ordered(self):
        har = generate_mock_har("x", "C", "A")
        self.tc.load_har(har)
        checklist_expected, checklist_actual = self.tc.check(ordered=False)
        self.assertEqual(checklist_expected, [True, False, True])
        self.assertEqual(checklist_actual, [False, True, True])

    def test_check_None(self):
        har = generate_mock_har("x")
        self.tc.load_har(har)
        checklist_expected, checklist_actual = self.tc.check()
        self.assertEqual(checklist_expected, [False, False, False])
        self.assertEqual(checklist_actual, [False])

    def test_check_no_http_log_loaded(self):
        with self.assertRaises(Exception):
            self.tc.check()


class test_Result(unittest.TestCase):
    def setUp(self) -> None:
        events = [{"dp": "A"}, {"dp": "B"}, {"dp": "C"}]
        tp = gaunit.TrackingPlan()
        tp.add_test_case("home_engie", events)
        self.tc = gaunit.TestCase("home_engie", tp)

    def test_was_successful(self):
        har = generate_mock_har("A", "B", "C")
        self.tc.load_har(har)
        r = self.tc.result()
        self.assertTrue(r.was_successful())

    def test_get_status_expected_events(self):
        har = generate_mock_har("A", "B")
        self.tc.load_har(har)
        r = self.tc.result()
        self.assertEqual(
            r.get_status_expected_events(),
            [
                {"event": {"dp": "A"}, "found": True},
                {"event": {"dp": "B"}, "found": True},
                {"event": {"dp": "C"}, "found": False},
            ],
        )

    def test_get_status_actual_events(self):
        har = generate_mock_har("A", "x")
        self.tc.load_har(har)
        r = self.tc.result()
        self.assertEqual(
            r.get_status_actual_events(),
            [
                {
                    "event": {"v": "1", "dp": "A"},
                    "expected": True,
                },
                {
                    "event": {"v": "1", "dp": "x"},
                    "expected": False,
                },
            ],
        )


if __name__ == "__main__":
    unittest.main()
